import 'dart:collection';
import 'dart:convert';
import 'dart:io';

import 'package:build/build.dart';
import 'package:fi18n/src/fi18n_impl.dart';

Builder arbToDartBuilder(BuilderOptions options) => ArbToDartBuilder(options);

Builder postBuilder(BuilderOptions options) => PostBuilder();

final _locales = <String,bool>{};
String _fi18nFileContent;
final _generatedLocalesDartFiles = <String>[];

class ArbToDartBuilder implements Builder {
  ArbToDartBuilder(this.options);

  final BuilderOptions options;

  @override
  final buildExtensions = const {
    '.arb': ['.fi18n']
  };

  @override
  Future<Function> build(BuildStep buildStep) async {
    // Each `buildStep` has a single input.
    var inputId = buildStep.inputId;

    final filePath = inputId.path.substring(0, inputId.path.lastIndexOf('/'));
    final defaultArbFilePathDir = 'lib/res';

    if (filePath != defaultArbFilePathDir) {
      throw Exception('Invalid .arb file path. add your files in $defaultArbFilePathDir');
    }

    String generatedClassName = 'S';
    final defaultLocale = options.config['default_locale'];

    print("DEFAULT LOCALE: $defaultLocale");


    final fileMeta = generateArbFileMetaFromArb(
      filename: inputId.pathSegments.last,
      defaultLocaleName: defaultLocale
    );

    _locales[fileMeta.localeName] = fileMeta.isDefault;

    if (fileMeta.localeName == defaultLocale) {
      // Generate temp file so that the next is executed
      final sClassContent = generateSClass(
        supportedLocales: _locales.keys.toList(),
        defaultArbContent: await buildStep.readAsString(inputId),
        generatedClassName: generatedClassName,
      );

      _fi18nFileContent = sClassContent;
      final languageDartFile = inputId.changeExtension('.fi18n');
      await buildStep.writeAsString(languageDartFile, sClassContent);
    }

    // .arb file content
    var arbContent = await buildStep.readAsString(inputId);

    final messages = json.decode(arbContent) as LinkedHashMap<String, dynamic>;

    final dartContent = generateLanguageScript(messages,
        locale: fileMeta.localeName, generatedClassName: generatedClassName);
    // save to use later
    _generatedLocalesDartFiles.add(dartContent);
    return Future.value();
  }
}

class PostBuilder implements Builder {

  @override
  final buildExtensions = const {
    '.fi18n': ['.dart']
  };

  @override
  Future<Function> build(BuildStep buildStep) async {
    String generatedClassName = 'S';
    final content = StringBuffer();

    content
      ..writeln('')
      ..writeln("import 'package:flutter/material.dart';")
      ..writeln("import 'package:flutter/foundation.dart';")
      ..writeln('')
      ..writeln(_fi18nFileContent);
    _generatedLocalesDartFiles.forEach((localeDartFile) {
      content.writeln(localeDartFile);
    });

    var defaultLocale = '';
    for (final i in _locales.keys) {
      if (_locales[i]) {
        defaultLocale = i;
      }
    }

    final delegateContent = generateDelegateScript(
      defaultArbContent: defaultLocale,
      objectClassName: generatedClassName,
      supportedLocales: _locales.keys.toList()
    );

    content.writeln(delegateContent);

    final fileName = 'fi18n.dart';
    final filePath = 'lib/res';
    final file = File('$filePath/$fileName');

    if (file.existsSync()) {
      await file.delete();
    }

    await file.writeAsString(content.toString(), mode: FileMode.write);
    file.create(recursive: true);
    return Future.value();
  }
}